# -*- coding: iso-8859-1 -*-
"""
Naive Bayesian classification

Vgl. Skript, "Anwendung: Ein Spamfilter"
"""

import sys, re, operator, os
from math import log


class Error(Exception):
	pass

class TrainError(Error):
	pass


def tokenize(tx):
	"""returns a list of individual words (defined by a regular expression)
	present in the text tx.
	"""
	return re.sub(r"(?L)[^\w\s']", " ", 
			re.sub("(?s)<[^>]*>", "", tx.lower())).split()

def getWordDistribution(tx):
	"""returns a dictionary for word frequencies in the text tx.
	"""
	freqDict = {}
	for word in tokenize(tx):
		freqDict[word] = freqDict.get(word, 0)+1
	return freqDict

def argmax(aList):
	maxVal, maxInd = aList[0]-1, 0
	for ind, val in enumerate(aList):
		if val==maxVal:
			raise Error, "Ambiguous list in argmax"
		if val>maxVal:
			maxVal, maxInd = val, ind
	return maxInd


class TrainableClassifier:
	"""A base class providing an interface to train text corpora.
	Child classes should provide a method learn that picks up the
	raw data (a mapping of words to dictionaries mapping
	the class to the frequency of the word in that class) to
	infer the classifier.  They can use self.numKlasses (which
	is the maximum class encountered) and should set isLearned
	to true.
	"""
	def __init__(self):
		self.rawData = {}
		self.isLearned = 0
		self.numKlasses = 0

	def train(self, klass, trainingData):
		"""adds the words in the text trainingData to the training set
		for class klass.
		"""
		self.numKlasses = max(self.numKlasses, klass+1)
		self.isLearned = 0
		for word, freq in getWordDistribution(trainingData).iteritems():
			kd = self.rawData.setdefault(word, {})
			kd[klass] = kd.get(klass, 0)+freq
	
	def classify(self, data):
		raise Error, "Abstract Classifier can't classify"


def specialNormalize(seq, minProb=0.01):
	"""normalizes seq such that sum(seq)=1.  To make up for estimation
	errors, we give non-observed cases a probablity of minProb and
	re-normalize.
	>>> specialNormalize([6, 3, 3])
	[0.5, 0.25, 0.25]
	>>> specialNormalize([5, 2, 0])
	[0.70721357850070721, 0.28288543140028288, 0.0099009900990099011]
	"""
	totalMass = float(sum(seq))
	seq = [val and val/totalMass or 0.01 
		for val in seq]
	totalMass = sum(seq)
	return [val/totalMass for val in seq]

class NaiveClassifier(TrainableClassifier):
	"""This class is a naive Bayesian Classifier.  To train it, feed
	text/class-pairs to the train method.  Call the learn method before
	calling classify to make it compute the conditional probabilities.
	We use a uniform prior, i.e. we do not try to derive priors from
	the distribution of training samples.
	"""
	def learn(self):
		"""estimates log(P(w|v)) for all words w and classes v shown
		to train() by almost-ML (almost since non-observed w for v lead
		to P(w|v)=minProb).
		"""
		self.probs = {}
		for word, wObs in self.rawData.iteritems():
			wDist = [0]*self.numKlasses
			for klass, freq in wObs.iteritems():
				wDist[klass] = freq
			wDist = specialNormalize(wDist)
			self.probs[word] = map(log, wDist)
		self.isLearned = 1

	def classify(self, tx):
		"""returns the most likely class for the text tx.  We ignore
		word frequency here and just look for presence or absence of
		words.  Unknown words are ignored.
		"""
		if not self.isLearned:
			raise Error, "You need to call learn() before classfiy"
		probs = [log(0.5)]*self.numKlasses
		for word in getWordDistribution(tx):
			wDist = self.probs.get(word, {})
			for klass in range(len(wDist)):
				probs[klass] += wDist[klass]
		return argmax(probs)



class GrahamClassifier(TrainableClassifier):
	"""A binary classifier written along the ideas put down by Paul
	Graham, http://www.paulgraham.com/spam.html
	Train by presenting spam (klass=1) and ham (klass=0) to train,
	then call learn to start the computation of the conditional
	probabilities, then call classify.
	"""
	def __init__(self, minProb=0.01, maxProb=0.99, minOccurrence=5,
			numBest=15):
		TrainableClassifier.__init__(self)
		self.minProb = minProb
		self.maxProb = maxProb
		self.minOccurrence = minOccurrence
		self.numBest = numBest

	def learn(self):
		self.posProbs = {}
		self.negProbs = {}
		for word, wObs in self.rawData.iteritems():
			negCount, posCount = wObs.get(0, 0)*2, wObs.get(1, 0)
			norm = float(negCount)+float(posCount)
			if norm<self.minOccurrence:
				continue
			self.posProbs[word] = max(self.minProb, min(self.maxProb,
				wObs.get(1, 0)/float(norm)))
		self.isLearned = 1

	def _computePosProb(self, evidence):
		"""returns a normalized probability for classifying a document
		with individual attribute probabilities as positive.
		"""
		scoredEvidence = [(abs(e[0]-0.5), e) for e in evidence]
		scoredEvidence.sort()
		bestEvidence = [pair[1][0] 
			for pair in scoredEvidence[-self.numBest:]]
		prod = reduce(operator.mul, bestEvidence)
		return prod/(prod+reduce(operator.mul, [1-e for e in bestEvidence]))

	def classify(self, data):
		"""returns the most likely class for data.  We count unknown
		words as having spam probablity 0.4.
		"""
		if not self.isLearned:
			raise Error, "You need to call learn() before classfiy"
		evidence = [(self.posProbs.get(w, 0.4), w)
				for w in getWordDistribution(data).keys()]
		posProb = self._computePosProb(evidence)
		return posProb>0.9
	

def _validateClassifier(classifier, srcDir, targetKlass):
	total, totalBad = 0, 0
	for fName in os.listdir(srcDir):
		targetPath = os.path.join(srcDir, fName)
		if classifier.classify(open(targetPath).read())!=targetKlass:
			print "Misclassified %s\n"%targetPath
			totalBad += 1
		total += 1
	return total, totalBad

def testClassifier(classifierClass):
	spamCorpus, hamCorpus = "spam.canned", "ham.canned"
	c = classifierClass()
	c.train(1, open(spamCorpus).read())
	c.train(0, open(hamCorpus).read())
	c.learn()
	print "Total Spam %d, false negatives %d"%_validateClassifier(
		c, "validation.spam", 1)
	print "Total Ham %d, false positives %d"%_validateClassifier(
		c, "validation.ham", 0)


def _test():
	import doctest, naiveBayesian
	return doctest.testmod(naiveBayesian)


if __name__=="__main__":
	_test()
	testClassifier(GrahamClassifier)
